#ifndef ANIM_H_
#define ANIM_H_

#include <fstream>
#include <iostream>
#include <allegro.h>
#include "helper.h"

using namespace std;

struct Frame {
  int tics;
  BITMAP *image;
};

struct Animation {
  int frame_count;
  int tic_sum;
  Frame *frame_array;
  
  Animation(string file) : tic_sum(0) {
    ifstream anim_file(file.c_str());
    anim_file >> frame_count;
    if (frame_count < 1) {
      allegro_message("Invalid frame count in animation file!");
      exit(-1);
    }
    frame_array = new Frame[frame_count];
    for (int i = 0; i < frame_count; i++) {
      string name;
      anim_file >> name;
      frame_array[i].image = Bitmap(name);
      anim_file >> frame_array[i].tics;
      if (frame_array[i].tics < 0) {
        allegro_message("Impossible tic count detected in animation file!");
        exit(-1);
      }
      tic_sum += frame_array[i].tics;
    }
  }
  
  ~Animation() {
    for (int i = 0; i < frame_count; i++) {
      destroy_bitmap(frame_array[i].image);
    }
    delete [] frame_array;
  }
  
  inline BITMAP *GetImage(int tics) {
    if (tics < 0) {
      tics += tic_sum * (1 + tics/tic_sum); // make positive
      assert(tics >= 0);
    }
    tics = tics % tic_sum;
    int i = -1;
    while (tics >= 0) {
      i++;
      tics -= frame_array[i].tics;
    }
    i = i % frame_count;
    
    return frame_array[i].image;
  }
  
  void Draw(int x, int y, int tics, float vx, BITMAP *dest) {
    if (vx < 0)
      draw_sprite_h_flip(dest, GetImage(tics), x, y);
    else
      draw_sprite(dest, GetImage(tics), x, y);
  }
  
  void Draw2(int x, int y, int tics, float vx, BITMAP *dest) { // same as draw1, but x, y designates the bottom middle instead of the top left of the image
    BITMAP *todraw = GetImage(tics);
    if (vx < 0) 
      draw_sprite_h_flip(dest, todraw, x - todraw->w/2, y - todraw->h);
    else
      draw_sprite(dest, todraw, x - todraw->w/2, y - todraw->h);
  }
};

#endif
