#include <allegro.h>
#include <string>
#include <list>
#include <fstream>
#include <cmath>
#include "vec2.h"
#include "anim.h"
#include "levels.h"
#include "helper.h"


#define LOGIC_RATE 60



enum State {STAND, CROUCH, SHOOT};

struct Player {
  Vec2f pos, lastpos;
  Vec2f vel;
  State state;
  bool jumpable;
  int last_bul_add;
  
  Animation *slow, *fast, *crouch, *shoot;
  
  Player(Animation *slow, Animation *fast, Animation *crouch, Animation *shoot) :
        slow(slow), fast(fast), crouch(crouch), shoot(shoot) {
    vel.mx = 0;
    vel.my = 0;
    jumpable = false;
    last_bul_add = anim_time;
  }
  
  
  bool CollideTest(Levels & levels, Vec2f p, int l) {
    for (float x = p.mx - 50; x < p.mx + 50; x += levels.middle->w - 1) {
      if (levels.Collide(Vec2f(x, p.my), l))
        return true;
    }
    return false;
  }
  
  void Collide(Levels &levels, int l) { 
    while (CollideTest(levels, pos, l)) {
      pos.my -= (int(pos.my) % levels.middle->h);
      vel.my = 0;
    }
    if (CollideTest(levels, pos + Vec2f(0,1), l))
      jumpable = true;
    else
      jumpable = false;
  }

  void Control(list<Bullet> &bullets, Levels &levels, int l) {
    float lr_accel = .2;
    
    if (key[KEY_RIGHT])
      vel.mx += lr_accel;
    
    
    if (key[KEY_LCONTROL] || key[KEY_UP] || key[KEY_DOWN]) {
      state = CROUCH;
    }
    else {
      if (state == CROUCH  && jumpable)
        vel.my = -5 + -5 * (fabs(vel.mx) / 10);
      if (key[KEY_SPACE] || key[KEY_LEFT]) {
        state = SHOOT;
        
        if (anim_time - last_bul_add >= 10) {
          vel.mx -= 1;
          Bullet b;
          b.vel = Vec2f(8+vel.mx,vel.my / 100);
          b.pos = pos + Vec2f(94, -250);
          b.time = anim_time;
          bullets.push_back(b);
          last_bul_add = anim_time;
        }
        if (vel.mx < -7)
          vel.mx = -7;
      }
      else {
        state = STAND;
      }
    }
    
    vel.my += .3;
    if (vel.my > 15)
      vel.my = 15;
    
    if (vel.mx > 10)
      vel.mx = 10;
    
    pos += vel;
    
    for (list<Bullet>::iterator i = bullets.begin(); i != bullets.end(); ++i) {
      i->pos += i->vel;
    }
    for (list<Bullet>::iterator i = bullets.begin(); i != bullets.end(); ++i) {
      if (anim_time - i->time > 15 || i->pos.mx > levels.Width(l)) {
        i = bullets.erase(i);
      }
    }
  }
  
  void Draw(BITMAP *buffer) {
    Animation *todraw = slow;
    if (state == STAND) {
      if (vel.mx > 2)
        todraw = fast;
    }
    else if (state == SHOOT) {
      todraw = shoot;
    }
    else if (state == CROUCH) {
      todraw = crouch;
    }
    
    todraw->Draw2(SCREEN_W/2, 3*SCREEN_H/4, anim_time, 0, buffer);
  }
};

void Splash(string file, FONT *f, int color = 0, int oldscore = -1, int diff = 0) {
  BITMAP *splash = Bitmap(file);
  draw_sprite(screen, splash, 0, 0);
  
  if (oldscore > -1) {
    BITMAP *buffer = create_bitmap(SCREEN_W, SCREEN_H);
    logic_pending = 0;
    
    float i = 0;
    while ( i < diff ) {
      while (logic_pending) {
        i+=.5;
        logic_pending--;
      }
      clear(buffer);
      draw_sprite(buffer, splash, 0, 0);
      textprintf_centre(buffer, f, SCREEN_W/2, SCREEN_H - 100, color, "Score: %d", oldscore + int(i));
      blit(buffer, screen, 0, 0, 0, 0, SCREEN_W, SCREEN_H);
    }
    destroy_bitmap(buffer);
  }
  
  int time = anim_time;
  
  
  while (keypressed()) {
    readkey();
  }
  
  while (!key[KEY_ENTER] && !key[KEY_ESC]) {
    if (anim_time - time > 30)
      textout_centre(screen, f, "Press Enter to Continue",  SCREEN_W/2, SCREEN_H - 50, color);
  }
  
  destroy_bitmap(splash);
}

struct Blood {
  Vec2f pos;
  int offset, time;
  
  void Draw(int x, int y, BITMAP *dest) {
    for (int iy = (int)pos.my - 5; iy < pos.my + 5; iy++) {
      line         (dest, 
                    int(sin((double)(iy + offset) / 25)*10 + pos.mx - x),  iy - y, 
                    int(sin((double)(iy+1 + offset) / 25)*10 + pos.mx - x), iy - y + 1, 
                    makecol(255,0,0));
    }
  }
};

int Kill(list<Bullet> &bullets, list<Target> &targets, list<Blood> &blood) {
  int killcount = 0;
  for (list<Target>::iterator t = targets.begin(); t != targets.end(); ++t) {
    BITMAP *image = t->walk->GetImage(anim_time);
    Vec2f p = t->pos - Vec2f(image->w/2, image->h);
    for (list<Bullet>::iterator b = bullets.begin(); b != bullets.end(); ++b) {
      if (b->pos.mx - p.mx < image->w && b->pos.mx - p.mx > 0 && b->pos.my - p.my < image->h && b->pos.my - p.my > 0) {
        int pix = getpixel(image, int(b->pos.mx - p.mx), int(b->pos.my - p.my));
        if (pix != makecol(255,0,255)) {
          Blood bl;
          for (int y = 0; y < image->h; y+=rand()%5+5) {
            for (int x = 0; x < image->w; x+=rand()%5+5) {
              int ex = x;
              if (t->vel.mx < 0) 
                ex = image->w - x - 1;
              if (rand()%400 && getpixel(image, ex, y) != makecol(255,0,255)) {
                bl.pos = Vec2f(x, y) + p;
                bl.offset = rand() % 20;
                bl.time = anim_time + rand() % 20;
                blood.push_back(bl);
              }
            }
          }
          killcount++;
          b = bullets.erase(b);
          t = targets.erase(t);
          break;
        }
      }
    }
  }
  
  return killcount;
}

enum Ending { NOT_YET, WIN, TIME_UP, ABYSS };

bool LevelToggle() {
  ifstream file("gamewon");
  string text;
  file >> text;
  if (text == "toggle")
    return true;
  
  else return false;
}

struct Score {
  string name;
  int score;
};

#define MAX_SCORES 6

void HighScores(int score, FONT *ourfont) {
  ifstream file("highscores");
  list<Score> scores;
  string name;
  int minscore = 0;
  while (file >> name) {
    Score s;
    s.name = name;
    file >> s.score;
    minscore = s.score;
    scores.push_back(s);
  }
  file.close();
  
  if (score > minscore || scores.size() < MAX_SCORES) {
    BITMAP *buffer = create_bitmap(SCREEN_W, SCREEN_H);
    
    text_mode(-1);
    
    char name[4];
    
    
    for (int ic = 0; ic < 3; ++ic) {
      
      name[ic] = 0;
      char c = 'A';
      while (key[KEY_ENTER]) {}
      while (!key[KEY_ENTER]) {
        clear_to_color(buffer, makecol(255,255,255));
        textout_centre(buffer, ourfont, "HIGH SCORE!  Enter 3 initials:", SCREEN_W/2, 30, makecol(0,0,0));
        textprintf_centre(buffer, ourfont, SCREEN_W/2, 250, makecol(255,0,0),  "%s%c", name, c);
        textout_centre(buffer, ourfont, "(press ENTER to accept a letter)", SCREEN_W/2, 450, makecol(0,0,0));
        textout_centre(buffer, ourfont, "(UP/DOWN to change the letter selection)", SCREEN_W/2, 500, makecol(0,0,0));
        blit(buffer, screen, 0, 0, 0, 0, SCREEN_W, SCREEN_H);
        int key = readkey();
        if (key >> 8 == KEY_UP) {
          c++;
          if (c > 'Z')
            c = 'A';
        }
        if (key >> 8 == KEY_DOWN) {
          c--;
          if (c < 'A') 
            c = 'Z';
        }
      }
      while (key[KEY_ENTER]) {}
      
      name[ic] = c;
    }
    name[3] = 0;
    int c = 0;
    bool swritten = false;
    clear_to_color(buffer, makecol(255,255,255));
    ofstream scorefile("highscores");
    textout_centre(buffer, ourfont, "The Hall of Infamy", SCREEN_W/2, 20, makecol(0,0,0));
    for (list<Score>::iterator i = scores.begin(); i != scores.end(); ++i, ++c) {
      if (c > MAX_SCORES)
        break;
      
      if (score > i->score && !swritten) {
        swritten = true;
        
        textprintf_centre(buffer, ourfont, SCREEN_W/2, c * 50 + 100, makecol(255,0,0), "%s: %d", name, score);
        scorefile << name << " ";
        scorefile << score << " " << endl;
        
        c++;
        if (c > MAX_SCORES)
          break;
      }
      
      textprintf_centre(buffer, ourfont, SCREEN_W/2, c * 50 + 100, 0, "%s: %d", i->name.c_str(), i->score);
      scorefile << i->name << " ";
      scorefile << i->score << " " << endl;
    }
    if (!swritten && c < MAX_SCORES) {
      swritten = true;
      
      textprintf_centre(buffer, ourfont, SCREEN_W/2, c * 50 + 100, makecol(255,0,0), "%s: %d", name, score);
      scorefile << name << " ";
      scorefile << score << " " << endl;
    }
    blit(buffer, screen, 0, 0, 0, 0, SCREEN_W, SCREEN_H);
    while (key[KEY_ENTER]&&!key[KEY_ESC]) {blit(buffer, screen, 0, 0, 0, 0, SCREEN_W, SCREEN_H);}
    while (!key[KEY_ENTER] && !key[KEY_ESC]) {blit(buffer, screen, 0, 0, 0, 0, SCREEN_W, SCREEN_H);}
    destroy_bitmap(buffer);
  }
}


int main() {
  
	
	// basic setup
  allegro_init();
  install_timer();
  install_mouse();
  install_keyboard();
  set_color_depth(16);
  if (set_gfx_mode(GFX_AUTODETECT_WINDOWED, 800, 600, 0, 0) != 0) {
    set_color_depth(15);
    if (set_gfx_mode(GFX_AUTODETECT_WINDOWED, 800, 600, 0, 0) != 0){
      allegro_message("Couldn't set graphics mode!");
      exit(1);
    }
  }
  
  char buf[1024];
  get_executable_name(buf, 1010);
  char *fname = (get_filename(buf));
  fname[0] = '/';
  fname[1] = 'd';
  fname[2] = 'a';
  fname[3] = 't';
  fname[4] = 'a';
  fname[5] = '/';
  fname[6] = 0;
  chdir(buf);
  
  DATAFILE* fontdat = load_datafile("font.dat");
  FONT* ourfont = (FONT*)fontdat[0].dat;
  text_mode(-1);
  if (!fontdat) {
    allegro_message("Failed to load font datafile");
    exit(1);
  }
  
  LOCK_VARIABLE(logic_pending);
  LOCK_VARIABLE(anim_time);
  LOCK_FUNCTION(logic_counter);
  LOCK_FUNCTION(anim_counter);
  install_int_ex(logic_counter, BPS_TO_TIMER(LOGIC_RATE));
  install_int_ex(anim_counter, BPS_TO_TIMER(30));
  
  BITMAP *buffer = create_bitmap(800, 600);
  
  Splash("splash.bmp", ourfont);
  
  
  Animation child("child.anim");
  Animation troll("troll.anim");
  Animation old("old.anim");
  Animation slow("player_slow.anim");
  Animation fast("Player_fast.anim");
  Animation crouch("Player_crouch.anim");
  Animation shoot("Player_shoot.anim");
  
  Player me(&slow, &fast, &crouch, &shoot);
  
  Levels levels("levels");
  
  
  list<Bullet> bullets;
  list<Blood> blood;
  
  bool leveltoggle = LevelToggle();
  
  int score = 0;
  int level;
  for (level = 0; level < levels.count; level++) {
    Ending level_end = NOT_YET;
    int killcount = 0;
    int start_time = anim_time;
    list<Target> targets = levels.Targets(level, &troll, &child, &old);
    me.pos = levels.StartPos(level);
    me.vel = Vec2f(0,0);
    bullets.empty();
    blood.empty();
    logic_pending = 0;
    bool level_skipped = false;
    while (!key[KEY_ESC]){
      
      while (logic_pending) {
        
        me.Control(bullets, levels, level);
        me.Collide(levels, level);
        levels.MoveTargets(targets, level);
        killcount += Kill(bullets, targets, blood);
        for (list<Blood>::iterator i = blood.begin(); i != blood.end(); ++i) {
          i->pos.my += .5 + (rand() % 3)/2;
          if (anim_time - i->time > 40 + rand()%20)
            i = blood.erase(i);
        }
        
        logic_pending--;
      }
      yield_timeslice();
      levels.Draw((int)me.pos.mx - SCREEN_W/2,(int) me.pos.my - 3*SCREEN_H/4, level, targets, buffer);
      me.Draw(buffer);
      for (list<Blood>::iterator i = blood.begin(); i != blood.end(); ++i) {
        i->Draw((int)me.pos.mx - SCREEN_W/2, (int) me.pos.my - 3*SCREEN_H/4, buffer);
      }
      for (list<Bullet>::iterator i = bullets.begin(); i != bullets.end(); ++i) { 
        i->Draw((int)me.pos.mx - SCREEN_W/2, (int) me.pos.my - 3*SCREEN_H/4, buffer);
      }
      textprintf(buffer, ourfont, 40, 20, makecol(128,0,0),  "Kill Count: %d / %d", killcount, levels.KillsRequired(level));
      textprintf_right(buffer, ourfont, SCREEN_W-40, 20, makecol(128,0,0),  "%d seconds remain", start_time/30 + levels.TimeLimit(level) - anim_time/30);
      
      if (killcount >= levels.KillsRequired(level)) {
        level_end = WIN;
        break;
      }
      else if (key[KEY_TAB] && leveltoggle) {
        level_end = WIN;
        level_skipped = true;
        break;
      }
      else if (start_time/30 + levels.TimeLimit(level) - anim_time/30 < 0) {
        level_end = TIME_UP;
        break;
      }
      else if (me.pos.mx < -100 || me.pos.mx > levels.Width(level) + 100 || me.pos.my > levels.Height(level) + 50) {
        level_end = ABYSS;
        break;
      }
      
      
      vsync();
      blit(buffer, screen, 0, 0, 0, 0, SCREEN_W, SCREEN_H);
      clear_to_color(buffer, makecol(255,255,255));
    }
    
    if (level_end == NOT_YET)
      break;
    if (level_end == TIME_UP) {
      Splash("time.bmp", ourfont);
      level--;
    }
    else if (level_end == ABYSS) {
      Splash("abyss.bmp", ourfont, makecol(255,0,0));
      level--;
    }
    else if (level_end == WIN) {
      int time_remaining = (start_time/30 + levels.TimeLimit(level) - anim_time/30);
      if (time_remaining < 0) time_remaining = 0;
      
      if (level+1 != levels.count)
        Splash("win.bmp", ourfont, makecol(0,0,0), score, time_remaining * !level_skipped);
      score += time_remaining * !level_skipped;
    }
  }
  if (level == levels.count) {
    ofstream file("gamewon");
    file << "toggle ";
    Splash("wingame.bmp", ourfont);
    
    HighScores(score, ourfont);
  }
  
  return 0;
} END_OF_MAIN()
