#ifndef _LEVELS_H_
#define _LEVELS_H_

#include <allegro.h>
#include <fstream>
#include <iostream>
#include <list>
#include "vec2.h"
#include "helper.h"
#include "anim.h"

using namespace std;

struct Level {
  BITMAP *content;
  int kills, time;
};

struct Target {
  Vec2f pos, vel;
  Animation *walk;
  int animoff;
};

struct Bullet {
  Vec2f pos, vel;
  int time;
  
  void Draw(int x, int y, BITMAP *dest) {
    for (int ix = (int)pos.mx - 10; ix < pos.mx + 10; ix++) {
      line         (dest, 
                    ix - x, int(sin((double)(ix+anim_time/5) / 20)*20 + pos.my - y), 
                    ix - x + 1, int(sin((double)(ix+1+anim_time/5) / 20)*20 + pos.my - y), 
                    makecol(255,0,0));
      
      line         (dest, 
                    ix - x, int(-sin((double)(ix+anim_time/5) / 20)*20 + pos.my - y), 
                    ix - x + 1, int(-sin((double)(ix+1+anim_time/5) / 20)*20 + pos.my - y), 
                    makecol(255,0,0));
    }
    circlefill(dest, (int)pos.mx - x, (int)pos.my - y, int(fabs(sin((double)pos.mx / 200)) * 5) + 2, makecol(255,0,0));
  }
};

struct Levels {
  BITMAP *left, *middle, *right;
  
  int count;
  Level *levels;
  
  Levels(string file) {
    ifstream data(file.c_str());
    data >> file;
    left = Bitmap(file);
    data >> file;
    middle = Bitmap(file);
    data >> file;
    right = Bitmap(file);
    
    data >> count;
    levels = new Level[count];
    for (int i = 0; i < count; i++) {
      data >> file;
      levels[i].content = Bitmap(file);
      data >> levels[i].kills;
      data >> levels[i].time;
    }
  }
  
  ~Levels() {
    for (int i = 0; i < count; i++) {
      destroy_bitmap(levels[i].content);
    }
    delete [] levels;
  }
  
  inline void Verify(int level) {
    if (level > count || level < 0) {
      allegro_message("I can't draw a level which does not exist!  And yet I'm asked to do it ... so I die.");
      exit(1);
    }
  }
  
  int Width(int level) {
    Verify(level);
    
    return levels[level].content->w * middle->w + SCREEN_W;
  }
  
  int KillsRequired(int level) {
    Verify(level);
    
    return levels[level].kills;
  }
  
  int TimeLimit(int level) {
    Verify(level);
    
    return levels[level].time;
  }
  
  int Height(int level) {
    Verify(level);
    
    return levels[level].content->h * middle->h + SCREEN_H;
  }
  
  bool Collide(Vec2f pos, int level) {
    Verify(level);
    
    if (int(pos.my) % middle->h == 0) 
      return 0;
    else
      return getpixel(levels[level].content, (int)pos.mx/middle->w, (int)pos.my/middle->h) == makecol(0,0,0);
  }
  
  list<Target> Targets(int level, Animation *troll, Animation *child, Animation *old) {
    Verify(level);
    
    cout << "Getting targets list" << endl;
    
    list<Target> toret;
    Target t;
    
    BITMAP* c = levels[level].content;
    for (int y = 0; y < c->h; y++) {
      for (int x = 0; x < c->w; x++) {
        int pix = getpixel(c, x, y);
        
        
        
        if (pix == makecol(0,255,0)) {
          t.pos = Vec2f(x * middle->w + middle->w/2, y * middle->h);
          t.vel = Vec2f( (rand() % 2) * 2 - 1, 0 );
          t.walk = troll;
          t.animoff = rand() % 20;
          toret.push_back(t);
          cout << "Adding troll to target list" << endl;
        }
        else if (pix == makecol(0,255,255)) {
          t.pos = Vec2f(x * middle->w + middle->w/2, y * middle->h);
          t.vel = Vec2f( (rand() % 2) * 2 - 1, 0 );
          t.walk = child;
          t.animoff = rand() % 20;
          toret.push_back(t);
          cout << "Adding child to target list" << endl;
        }
        else if (pix == makecol(255, 255, 0)) {
          t.pos = Vec2f(x * middle->w + middle->w/2, y * middle->h);
          t.vel = Vec2f( 0, 0 );
          t.walk = old;
          t.animoff = rand() % 20;
          toret.push_back(t);
          cout << "Adding old man to target list" << endl;
        }
      }
    }
    
    return toret;
  }
  
  void MoveTargets(list<Target> &targets, int level) {
    Verify(level);
    
    for (list<Target>::iterator i = targets.begin(); i != targets.end(); ++i) {
      i->vel.my += .3;
      if (i->vel.my > 15)
        i->vel.my = 15;
      
      if (!Collide(i->pos + Vec2f(i->vel.mx, 1), level) && Collide(i->pos + Vec2f(0, 1), level))
        i->vel.mx = -i->vel.mx;
      
      i->pos += i->vel;
      //cout << "Moving guy at " << i->pos.mx << ", " << i->pos.my << " by " << i->vel.mx << ", " << i->vel.my << endl;
      
      while (Collide(i->pos, level)) {
        i->pos.my -= (int(i->pos.my) % middle->h);
      }
    }
  }
  
  Vec2f StartPos(int level) {
    Verify(level);
    
    BITMAP *content = levels[level].content;
    for (int y = 0; y < content->h; y++) {
      for (int x = 0; x < content->w; x++) {
        if (getpixel(content, x, y) == makecol(255,0,0))
          return Vec2f(x * middle->w, y * middle->h);
      }
    }
    
    allegro_message("Level is missing starting position; cannot play!");
    exit(1);
  }
  
  void Draw(int x, int y, int level, list<Target> targets, BITMAP *dest) {
    Verify(level);
    
    BITMAP *content = levels[level].content;
    int tile_width = middle->w;
    int tile_height = middle->h;
    
    if (x + SCREEN_W > tile_width * (content->w - 3)) {
      rectfill(dest, tile_width * (content->w - 3) - x, 0, SCREEN_W, SCREEN_H, makecol(0,0,0));
    }
    if (x < tile_width * 1) {
      rectfill(dest, 0, 0, tile_width * (1) - x, SCREEN_H, makecol(0,0,0));
    }
    if (y + SCREEN_H > tile_height * (content->h)) {
      rectfill(dest, 0, tile_height * (content->h) - y, SCREEN_W, SCREEN_H, makecol(0,0,0));
    }
    
    for (int iy = y/tile_height; iy < content->h && iy < y/tile_height + SCREEN_H/tile_height + 1; iy++) {
      for (int ix = x/tile_width; ix < content->w && ix < x/tile_width + SCREEN_W/tile_width + 2; ix++) {
        int pix = getpixel(content, ix, iy);
        if (pix == -1) {
          continue;
        }
        else if (pix == makecol(0,0,0)) {
          if (getpixel(content, ix-1, iy) == pix) {
            if (getpixel(content, ix+1, iy) == pix) {
              draw_sprite(dest, middle, -x + tile_width * ix, -y + tile_height * iy);
            }
            else {
              draw_sprite(dest, right, -x + tile_width * ix, -y + tile_height * iy - (right->h - tile_height));
            }
          }
          else if (getpixel(content, ix+1, iy) == pix) {
            draw_sprite(dest, left, -x + tile_width * ix, -y + tile_height * iy);
          }
        }
      }
    }
    
    for (list<Target>::iterator i = targets.begin(); i != targets.end(); ++i) {
      (i->walk)->Draw2((int)i->pos.mx - x, (int)i->pos.my - y, anim_time + i->animoff, i->vel.mx, dest);
    }
  }
};

#endif
